#
#
# changes version for PHOENIX WITHOUT HDF5 plugin to be revised/renamed...
#
#

import enum
import os
import threading
import time

from bec_lib.logger import bec_logger
from ophyd import Component as Cpt
from ophyd import Device, EpicsSignal, EpicsSignalRO, EpicsSignalWithRBV
from ophyd.mca import EpicsMCARecord
from ophyd_devices.interfaces.base_classes.psi_detector_base import (
    CustomDetectorMixin,
    PSIDetectorBase,
)


from ophyd.mca import EpicsDXP

from ophyd_devices.devices.dxp import xMAP, EpicsMCARecord

#from phoenix_bec.devices.dxp_loc import xMAP, EpicsMCARecord

from ophyd_devices.devices.areadetector.plugins import HDF5Plugin_V35 as HDF5Plugin

logger = bec_logger.logger
# bec_logger.level = bec_logger.LOGLEVEL.TRACE
bec_logger.level = bec_logger.LOGLEVEL.INFO


class XMAPError(Exception):
    """Base class for exceptions in this module."""


class XMAPTimeoutError(XMAPError):
    """Raised when the XMAP does not respond in time."""


class DetectorState(enum.IntEnum):
    """Detector states for XMAP detector"""

    DONE = 0
    ACQUIRING = 1


class TriggerSource(enum.IntEnum):
    """Trigger source for XMAP detector"""

    USER = 0
    GATE = 1
    SYNC = 2


class MappingSource(enum.IntEnum):
    """Mapping source for XMAP detector"""

    SPECTRUM = 0
    MAPPING = 1


class XMAPSetup(CustomDetectorMixin):
    """
    XMAP setup class for phoenix

    Parent class: CustomDetectorMixin

    """

    def __init__(self, *args, parent: Device = None, **kwargs) -> None:

        super().__init__(*args, parent=parent, **kwargs)
        self._lock = threading.RLock()

    def on_init(self) -> None:
        """Initialize XMAP detector"""
        self.initialize_default_parameter()
        self.initialize_detector()
        self.initialize_detector_backend()

    def initialize_default_parameter(self) -> None:
        """
        Set default parameters for XMAP

        This will set:
        - readout (float): readout time in seconds
        - value_pixel_per_buffer (int): number of spectra in buffer of XMAP

        """
        # self.parent.value_pixel_per_buffer = 20
        # self.update_readout_time()
        pass

    def update_readout_time(self) -> None:
        """Set readout time for Eiger9M detector"""
        """
        readout_time = (
            self.parent.scaninfo.readout_time
            if hasattr(self.parent.scaninfo, "readout_time")
            else self.parent.MIN_READOUT
        )
        self.parent.readout_time = max(readout_time, self.parent.MIN_READOUT)
        """

    def initialize_detector(self) -> None:
        """Initialize XMAP detector"""
        """
        self.stop_detector()
        self.stop_detector_backend()
        self.set_trigger(
            mapping_mode=MappingSource.MAPPING, trigger_source=TriggerSource.GATE, ignore_gate=0
        )
        # 1 Realtime
        self.parent.preset_mode.put(1)
        # 0 Normal, 1 Inverted
        self.parent.input_logic_polarity.put(0)
        # 0 Manual 1 Auto
        self.parent.auto_pixels_per_buffer.put(0)
        # Sets the number of pixels/spectra in the buffer
        self.parent.pixels_per_buffer.put(self.parent.value_pixel_per_buffer)
        """
        pass

    def initialize_detector_backend(self) -> None:
        """
        Initialize the detector backend for XMAP.
        currently no function
        """

        pass

        # ----------------------------------------------------------------------
        # self.parent.hdf5.enable.put(1)
        # file location of h5 layout for cSAXS
        # self.parent.hdf5.xml_file_name.put("layout.xml")
        # TODO Check if lazy open is needed and wanted!
        # self.parent.hdf5.lazy_open.put(1)
        # self.parent.hdf5.temp_suffix.put("")
        # size of queue for number of spectra a time.sleep(0.05)llowed in the buffer, if too small at high throughput, data is lost
        # self.parent.hdf5.queue_size.put(2000)
        # Segmentation into Spectra within EPICS, 1 is activate, 0 is deactivate
        # self.parent.nd_array_mode.put(1)

    def on_stage(self) -> None:
        """
        Prepare detector and backend for acquisition
        """

        # staging for XMAP as used in FDA this is for step scan
        # make sure that detector is ready to measure

        self.parent.stop_all.set(1)
        time.sleep(0.05)
        self.parent.collect_mode.set(0)
        time.sleep(0.05)
        self.parent.apply.set(0)
        time.sleep(0.05)
        self.parent.preset_mode.set(0)

        time.sleep(0.05)

        # self.prepare_detector()
        # self.prepare_data_backend()
        # self.publish_file_location(done=False, successful=False)
        # self.arm_acquisition() time.sleep(0.05)

    def prepare_detector(self) -> None:
        """Prepare detector for acquisition.. originally called from on stage """
        pass
        """
        self.set_trigger(
            mapping_mode=MappingSource.MAPPING, trigger_source=TriggerSource.GATE, ignore_gate=0
        )
        self.parent.preset_real.put(self.parent.scaninfo.exp_time)
        self.parent.pixels_per_run.put(
            int(self.parent.scaninfo.num_points * self.parent.scaninfo.frames_per_trigger)
        )
        """

    def prepare_data_backend(self) -> None:
        """Prepare data backend for acquisition.. originally called from on_stage """
        pass
        """ --------------------------------------------------------------
        self.parent.filepath.set(
            self.parent.filewriter.compile_full_filename(f"{self.parent.name}.h5")
        ).wait()
        file_path, file_name = os.path.split(self.parent.filepath.get())
        self.parent.hdf5.file_path.put(file_path)
        self.parent.hdf5.file_name.put(file_name)
        self.parent.hdf5.file_template.put("%s%s")
        self.parent.hdf5.num_capture.put(
            int(self.parent.scaninfo.num_points * self.parent.scaninfo.frames_per_trigger)
        )
        self.parent.hdf5.file_write_mode.put(2)
        # Reset spectrum counter in filewriter, used for indexing & identifying missing triggers
        self.parent.hdf5.array_counter.put(0)
        # Start file writing
        self.parent.hdf5.capture.put(1)
        """

    def arm_acquisition(self) -> None:
        """Arm detector for acquisition called fron on stage """
        pass
        """
        self.parent.start_all.put(1)
        signal_conditions = [
            (
                lambda: self.parent.acquiring.read()[self.parent.acquiring.name]["value"],
                DetectorState.ACQUIRING,
            )
        ]
        if not self.wait_for_signals(
            signal_conditions=signal_conditions,
            timeout=self.parent.TIMEOUT_FOR_SIGNALS,
            check_stopped=True,
            all_signals=False,
        ):
            raise XMAPTimeoutError(
                f"Failed to arm the acquisition. Detector state {signal_conditions[0][0]}"
            )
        """

    def on_unstage(self) -> None:
        """Unstage detector and backend"""
        pass

    def on_complete(self) -> None:
        """Complete detector and backend"""
        # ------------------------------------------------------------------
        # self.finished(timeout=self.parent.TIMEOUT_FOR_SIGNALS)
        # self.publish_file_location(done=True, successful=True)
        w = 9

    def on_stop(self) -> None:
        """Stop detector and backend"""
        self.stop_detector()

        # self.stop_detector_backend()

    def stop_detector(self) -> None:
        """Stops detector called from on _sttop"""
        self.parent.stop_all.put(1)
        self.parent.erase_all.put(1)

        # -------------------------------------------------------------------
        # signal_conditions = [
        #    (lambda: self.parent.acquiring.read()[self.parent.acquiring.name]["value"], DetectorState.DONE)
        # ]stage2 = StageXY(prefix='X07MB',name='-ES-MA1', name='stage2')
        #    timeout=self.parent.TIMEOUT_FOR_SIGNALS - self.parent.TIMEOUT_FOR_SIGNALS // 2,
        #    all_signals=False,
        # ):
        #    # Retry stop detector and wait for remaining time
        #    raise XMAPTimeoutError(
        #        f"Failed to stop detector, timeout with state {signal_conditions[0][0]}"
        #    )

    def stop_detector_backend(self) -> None:
        """Stop the detector backend"""
        self.parent.hdf5.capture.put(0)


    def finished(self, timeout: int = 5) -> None:
        """Check if scan finished succesfully called fon on _complete """
        with self._lock:
            total_frames = int(
                self.parent.scaninfo.num_points * self.parent.scaninfo.frames_per_trigger
            )
            signal_conditions = [
                (self.parent.dxp.current_pixel.get, total_frames),
                #     (self.parent.hdf5.array_counter.get, total_frames), ---------------------
            ]
            if not self.wait_for_signals(
                signal_conditions=signal_conditions,
                timeout=timeout,
                check_stopped=True,
                all_signals=True,
            ):
                logger.debug(
                    f"XMAP missed a trigger: received trigger {self.parent.dxp.current_pixel.get()},"
                    f" send data {self.parent.hdf5.array_counter.get()} from total_frames"
                    f" {total_frames}"
                )
            self.stop_detector()

            self.stop_detector_backend()

    def set_trigger(
        self, mapping_mode: MappingSource, trigger_source: TriggerSource, ignore_gate: int = 0
    ) -> None:
        """
        Set triggering mode for finisheddetector

        Args:
            mapping_mode (MappingSource): Mapping mode for the detector
            trigger_source (TriggerSource): Trigger source for the detector, pixel_advance_signal
            ignore_gate (int): Ignore gate from TTL signal; defaults to 0

        """
        pass

        """
        mapping = int(mapping_mode)
        trigger = trigger_source
        self.parent.collect_mode.put(mapping)
        self.parent.pixel_advance_mode.put(trigger)
        self.parent.ignore_gate.put(ignore_gate)
        """

class XMAPPhoenix(PSIDetectorBase, xMAP):
    """
    XMAP 4-element detector for phoenix
        custom_prepare_cls (XMAPSetup)          : Custom detector setup class for PHOENIX,
                                                  inherits from CustomDetectorMixin
                                                  in __init__ of PSIDetecor base
        PSIDetectorBase.set_min_readout (float) : Minimum readout time for the detector
        dxp<n> (EpicsDXPXMAP)                   : DXP parameters for XMAP detector
        mca<n> (EpicsMCARecord)                 : MCA parameters for XMAP detector
        hdf5 (XMAPHDF5Plugins)                  : HDF5 parameters for XMAP detector
        MIN_READOUT (float)                     : Minimum readout time for the detector
    """

    # Specify which functions are revealed to the user in BEC client
    USER_ACCESS = ["describe", "ep"]

    # specify Setup class
    custom_prepare_cls = XMAPSetup
    # specify minimum readout time for detector
    MIN_READOUT = 3e-3
    TIMEOUT_FOR_SIGNALS = 5

    # for fast loading uncomment dxp and mca for now
    # dxp1 = Cpt(EpicsDXP, "dxp1:")
    # dxp2 = Cpt(EpicsDXP, "dxp2:")
    # dxp3 = Cpt(EpicsDXP, "dxp3:")
    # dxp4 = Cpt(EpicsDXP, "dxp4:")

    # mca1 = Cpt(EpicsMCARecord, "mca1")
    # mca2 = Cpt(EpicsMCARecord, "mca2")
    # mca3 = Cpt(EpicsMCARecord, "mca3")
    # mca4 = Cpt(EpicsMCARecord, "mca4")

    hdf5 = Cpt(HDF5Plugin, "HDF1:")


    # for compatibility and mapping ruke EPICS --> bec

    preset_real = xMAP.preset_real_time
    preset_live = xMAP.preset_live_time

    # create some easy to find names for frequently used channels

    ph_erase_start = xMAP.erase_start
    ph_start_all = xMAP.start_all
    ph_stop_all = xMAP.stop_all
    ph_erase_all = xMAP.erase_all
    ph_collect_mode = xMAP.collect_mode
    ph_preset_mode = xMAP.preset_mode
    ph_preset_real = xMAP.preset_real_time
    ph_preset_live = xMAP.preset_live_time
    ph_elapsed_real = xMAP.elapsed_real
    ph_elapsed_live = xMAP.elapsed_live


